@@ -28,7 +28,7 @@ struct some_alloc
     void allocate(size_t);
 };
 
-int main(int, char**)
+TEST_CONSTEXPR_CXX20 bool tests()
 {
     {
         typedef std::vector<MoveOnly> C;
@@ -49,5 +49,14 @@ int main(int, char**)
     }
 #endif // _LIBCPP_VERSION
 
-  return 0;
+    return true;
+}
+
+int main(int, char**)
+{
+    tests();
+#if TEST_STD_VER > 17
+    static_assert(tests());
+#endif
+    return 0;
 }
